import astropy.units as u
import pytest

from demcmc import EmissionLine, TempBins
from demcmc.emission import ContFuncDiscrete


@pytest.fixture
def cont_func() -> ContFuncDiscrete:
    temps = [1, 2, 3, 4, 5] * u.MK
    values = [0.1, 0.3, 0.5, 0.7, 0.5] * u.cm**5 / u.K
    return ContFuncDiscrete(temps, values)


class TestContFuncDiscrete:
    def test_binned_error(self, cont_func):
        tbins = TempBins([0, 1, 3, 4] * u.MK)
        with pytest.raises(
            ValueError,
            match="The following bin edges in temp_bins are missing from the contribution function temperature coordinates",
        ):
            cont_func.binned(tbins)

    def test_binned(self, cont_func):
        tbins = TempBins([1, 3, 5] * u.MK)
        assert u.allclose(cont_func.binned(tbins), [0.3, 0.6] * u.cm**5 / u.K)


def test_emission_line(cont_func):
    line = EmissionLine(cont_func, name="my_line")
    assert line.name == "my_line"



import astropy.units as u
import numpy as np
import pytest

from demcmc.dem import TempBins


@pytest.fixture
def temp_bins() -> TempBins:
    return TempBins([1, 2, 4] * u.K)


class TestTempBins:
    def test_widths(self, temp_bins: TempBins) -> None:
        assert u.allclose(temp_bins.bin_widths, [1, 2] * u.K)

    def test_len(self, temp_bins: TempBins) -> None:
        assert len(temp_bins) == 2

    def test_wrong_units(self) -> None:
        with pytest.raises(u.UnitsError):
            TempBins([1, 2] * u.cm)

    def test_setters(self, temp_bins: TempBins) -> None:
        with pytest.raises(RuntimeError):
            temp_bins.edges = [2, 3] * u.K

    def test_centers(self, temp_bins: TempBins) -> None:
        assert np.all(temp_bins.bin_centers == [1.5, 3] * u.K)

    def test_iter_bins(self, temp_bins: TempBins) -> None:
        expected = [(1, 2) * u.K, (2, 4) * u.K]
        for b, e in zip(temp_bins.iter_bins(), expected):
            assert np.all(b == e)



"""
Structures for storing and working with emission lines.
"""
import functools
from abc import ABC, abstractmethod, abstractproperty
from dataclasses import dataclass
from typing import Any, List, Optional

import astropy.units as u
import numpy as np
import pandas as pd
from matplotlib.axes import Axes

from demcmc.dem import BinnedDEM, TempBins
from demcmc.units import u_cont_func

__all__ = [
    "ContFunc",
    "ContFuncGaussian",
    "ContFuncDiscrete",
    "EmissionLine",
    "plot_emission_loci",
]


class ContFunc(ABC):
    """
    A contribution function.
    """

    @abstractmethod
    def binned(self, temp_bins: TempBins) -> u.Quantity:
        """
        Get contribution function averaged over a number of temperature bins.

        Parameters
        ----------
        temp_bins : TempBins
            Temperature bins to get contribution function at.

        Returns
        -------
        astropy.units.Quantity
            Contribution function at given temperature bins.
        """

    @abstractmethod
    def _binned_arr(self, temp_bins: TempBins) -> np.ndarray:
        """
        Get contribution function averaged over a number of temperature bins.

        Should be returned as a bare array with units of Kelvin.

        Parameters
        ----------
        temp_bins : TempBins
            Temperature bins to get contribution function at.

        Returns
        -------
        astropy.units.Quantity
            Contribution function at given temperature bins.
        """

    @abstractproperty
    def temps(self) -> TempBins:
        """
        Default bins at which to evaluate the contribution function when plotting.
        """


class ContFuncGaussian(ContFunc):
    """
    A contribution function with a Gaussian profile.

    Parameters
    ----------
    center : u.Quantity
        Center of the Gaussian contribution function.
    width : u.Quantity
        Width of the Gaussian contribution function.
    """

    def __init__(self, center: u.Quantity, width: u.Quantity):
        self.width = width
        self.center = center

        self._width_MK = self.width.to_value(u.MK)
        self._center_MK = self.center.to_value(u.MK)

    def _binned_arr(self, temp_bins: TempBins) -> np.ndarray:
        """
        Get contribution function without units.
        """
        bins = temp_bins.bin_centers.to_value(u.MK)
        return (
            np.exp(
                -(((self._center_MK - temp_bins._bin_centers_MK) / self._width_MK) ** 2)
            )
        ) * 1e6

    def binned(self, temp_bins: TempBins) -> u.Quantity:
        """
        Get contribution function.

        Parameters
        ----------
        temp_bins : TempBins
            Temperature bins to get contribution function at.

        Returns
        -------
        astropy.units.Quantity
            Contribution function at given temperature bins.
        """
        return self._binned_arr(temp_bins) * u_cont_func

    @property
    def temps(self) -> TempBins:
        """
        Temperature bins with 100 samples.
        """
        edges = np.linspace(
            self.center - 3 * self.width, self.center + 3 * self.width, 100
        )
        bins: TempBins = TempBins(edges)
        return bins


class ContFuncDiscrete(ContFunc):
    """
    A pre-computed contribution function defined at temperature values.

    Parameters
    ----------
    temps : u.Quantity
        Temperature values of samples.
    values : u.Quantity
        Contribution function values.
    name : str
        Name for the contribution function.
    """

    def __init__(
        self,
        temps: u.Quantity[u.K],
        values: u.Quantity[u.cm**5 / u.K],
        *,
        name: Optional[str] = None,
    ):
        if temps.ndim != 1:
            raise ValueError("temps must be a 1D quantity")
        if values.ndim != 1:
            raise ValueError("values must be a 1D quantity")
        if temps.size != values.size:
            raise ValueError("Temperatures and values must be the same size")

        self._temps = temps
        self._values = values
        self._name = name
        self._hash = id(self)

    def __repr__(self) -> str:
        return f"ContFuncDiscrete(name={self._name}, len(temps)={len(self._temps)})"

    def __str__(self) -> str:
        if self._name is None:
            return f"Discrete contribution function sampled at {len(self._temps)} temperatures"
        else:
            return f"{self._name} contribution function sampled at {len(self._temps)} temperatures"

    @property
    def temps(self) -> u.Quantity[u.K]:
        """
        Temperatures of contribution function samples.
        """
        return self._temps

    @temps.setter
    def temps(self, val: Any) -> None:
        """
        Raises an error.
        """
        raise RuntimeError("ContFuncDiscrete instances are immutable")

    @property
    def values(self) -> u.Quantity[u.cm**5 / u.K]:
        """
        Contribution function values.
        """
        return self._values

    @values.setter
    def values(self, val: Any) -> None:
        """
        Raises an error.
        """
        raise RuntimeError("ContFuncDiscrete instances are immutable")

    def __hash__(self) -> int:
        return self._hash

    def _check_bin_edges(self, temp_bins: TempBins) -> None:
        missing_ts = []
        for t in temp_bins.edges:
            if not np.any(u.isclose(t, self._temps, atol=1 * u.K, rtol=0)):
                missing_ts.append(t)
        if len(missing_ts):
            raise ValueError(
                f"The following bin edges in temp_bins are missing from the contribution function temperature coordinates: {missing_ts}"
            )

    # Can change to just .cache() when Python 3.8 support dropped
    @functools.lru_cache(maxsize=None)
    def _binned_arr(self, temp_bins: TempBins) -> np.ndarray:
        self._check_bin_edges(temp_bins)

        df = pd.DataFrame(
            {
                "Temps": self.temps.to_value(u.K),
                "values": self.values.to_value(u.cm**5 / u.K),
            }
        )
        df = df.set_index("Temps")
        df["Groups"] = pd.cut(
            df.index, temp_bins.edges.to_value(u.K), include_lowest=True
        )
        means = df.groupby("Groups").mean()["values"].values
        return means

    # Can change to just .cache() when Python 3.8 support dropped
    @functools.lru_cache(maxsize=None)
    def binned(self, temp_bins: TempBins) -> u.Quantity[u.cm**5 / u.K]:
        """
        Get contribution function.

        Parameters
        ----------
        temp_bins : TempBins
            Temperature bins to get contribution function at.

        Returns
        -------
        astropy.units.Quantity
            Contribution function at given temperature bins.
        """
        return self._binned_arr(temp_bins) * u.cm**5 / u.K


@dataclass
class EmissionLine:
    """
    A single emission line.

    Parameters
    ----------
    cont_func : ContFunc
        Contribution function.
    intensity_obs : float
        Observed intensity.
    sigma_intensity_obs : float
        Uncertainty in observed intensity.
    name : str
        Emission line name.
    """

    cont_func: ContFunc
    intensity_obs: Optional[float] = None
    sigma_intensity_obs: Optional[float] = None
    name: Optional[str] = None

    def I_pred(self, dem: BinnedDEM) -> u.Quantity:
        """
        Calculate predicted intensity of a given line.

        Parameters
        ----------
        dem : BinnedDEM
            DEM.

        Returns
        -------
        astropy.units.Quantity
            Predicted intensity.
        """
        cont_func = self.cont_func._binned_arr(dem.temp_bins)
        return np.sum(cont_func * dem._values_arr * dem.temp_bins._bin_widths_arr)

    def _I_pred(self, temp_bins: TempBins, dem_values: np.ndarray) -> np.ndarray:
        """
        Same as above, but not using quantities for speed.
        """
        cont_func = self.cont_func._binned_arr(temp_bins)
        return np.sum(cont_func * dem_values * temp_bins._bin_widths_arr)


def plot_emission_loci(lines: List[EmissionLine], ax: Axes, **kwargs: Any) -> None:
    """
    Plot emission loci for a set of observed emission lines.

    Parameters
    ----------
    lines : list[EmissionLine]
        Lines to plot.
    ax : matplotlib.axes.Axes
        Axes to plot on.
    **kwargs : dict
        Keyword arguments are passed to :meth:`~matplotlib.axes.Axes.stairs`.

    Notes
    -----
    Currently only works with lines that have a `ContFuncDiscrete`
    contribution function.
    """
    kwargs.setdefault("color", "k")
    # Plot emission loci
    for line in lines:
        if not isinstance(line.cont_func, ContFuncDiscrete):
            continue
        tbins = TempBins(line.cont_func.temps)
        cont_func = line.cont_func.binned(tbins)
        with np.errstate(over="ignore", divide="ignore"):
            locus = line.intensity_obs / cont_func / tbins.bin_widths
        ax.stairs(locus, tbins.edges, **kwargs)



import astropy.units as u
import numpy as np
import pytest
import xarray as xr

from demcmc.dem import DEMOutput
from demcmc.emission import EmissionLine, TempBins
from demcmc.io import load_cont_funcs
from demcmc.mcmc import predict_dem_emcee
from demcmc.sample_data import fetch_sample_data


@pytest.fixture
def cont_funcs():
    _, cont_func_path = fetch_sample_data()
    return load_cont_funcs(cont_func_path)


@pytest.fixture
def line_intensities():
    line_intensities_path, _ = fetch_sample_data()
    return xr.load_dataarray(line_intensities_path)


@pytest.fixture
def lines(line_intensities, cont_funcs):
    line_intensities, cont_funcs

    lines = []
    for line in line_intensities.coords["Line"].values:
        cont_func = cont_funcs[line]
        intensity = line_intensities.loc[line, :]

        line = EmissionLine(
            cont_func,
            intensity_obs=intensity.loc["Intensity"].values,
            sigma_intensity_obs=intensity.loc["Error"].values,
        )
        lines.append(line)

    return lines


def test_mcmc(lines, tmpdir):
    """
    Smoke test of the MCMC run. Does NOT check values are correct or sensible.
    """
    temp_bins = TempBins(10 ** np.arange(5.6, 6.8, 0.1) * u.K)
    dem_result = predict_dem_emcee(lines, temp_bins, nsteps=1, nwalkers=30)
    assert isinstance(dem_result, DEMOutput)

    save_path = tmpdir / "result.nc"
    dem_result.save(save_path)
    assert save_path.exists()

    DEMOutput.load(save_path)



"""
Calculating a DEM from synthetic data
=====================================
This page contains a tutorial, stepping the user through estimating a DEM.
We don't use any real data here, but instead a series of fake lines that have Gaussian contribution functions.
"""
##################################################################
# Start by importing the required modules
import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
from astropy.visualization import quantity_support

from demcmc.dem import BinnedDEM, TempBins
from demcmc.emission import ContFuncGaussian, EmissionLine
from demcmc.mcmc import predict_dem_emcee

quantity_support()

##################################################################
# To start with we'll create a set of syntemtic emission lines.
# These are created with their line center and line width, which
# define their Gaussian contribution functions.
lines = []

# 20 lines, evenly spaced from 1 MK -> 2 MK
line_centers = np.linspace(1, 2, 21) * u.MK
line_width = 0.1 * u.MK

for line_center in line_centers:
    cont_func = ContFuncGaussian(line_center, line_width)
    lines.append(EmissionLine(cont_func))

# 600 temperature bins, linearly spaced from 0.5 -> 2.5 MK
temp_bins = TempBins(np.linspace(0.5, 2.5, 601) * u.MK)

fig, ax = plt.subplots()
for line in lines:
    ax.stairs(line.cont_func.binned(temp_bins), temp_bins.edges)

ax.set_title("Line contribution functions")

##################################################################
# Now lets create a 'fake' input DEM. We will use this to simulate
# the intensities that each of these lines would emit.
#
# This input DEM is calculated on a very coarse temperature grid,
# mainly to make the DEM inversion later much faster for this
# example.
coarse_temps = TempBins(np.linspace(1, 2, 6) * u.MK)
dem_in = np.exp(-(((coarse_temps.bin_centers - 1.2 * u.MK) / (0.2 * u.MK)) ** 2))
dem_in = BinnedDEM(coarse_temps, dem_in * u.cm**-5)

fig, axs = plt.subplots(nrows=2, sharex=True)
ax = axs[0]
ax.stairs(dem_in.values, coarse_temps.edges)
ax.set_title("Input DEM")

ax = axs[1]
for line in lines:
    ax.stairs(line.cont_func.binned(temp_bins), temp_bins.edges)

ax.set_ylim(bottom=0)
ax.set_title("Line contribution functions")

##################################################################
# Now lets use this DEM and the line contribution functions to
# simulate the intensity that each line would observe.
for line in lines:
    I_pred = line.I_pred(dem_in)
    line.intensity_obs = I_pred
    # Set error to 1/10th of observation
    line.sigma_intensity_obs = I_pred / 10


centers = u.Quantity([line.cont_func.center for line in lines])
intensities = u.Quantity([line.intensity_obs for line in lines])

fig, ax = plt.subplots(constrained_layout=True)
ax.scatter(centers, intensities)
ax.set_title("Simulated line intensities")
ax.set_xlabel("Line contribution function center / MK")

##################################################################
# Now pretend you didn't see the input DEM above! The problem we
# want to solve is:
#
#   Given the line contribution functions, and the observed
#   intensity in each line, what was the original DEM?
#
# To do this we use the ``predict_dem_emcee`` function to
# do a DEM inversion. This takes the list of lines we created
# earlier, and the temperature bins that we want to estimate
# the DEM in.

# Run DEM inversion
dem_result = predict_dem_emcee(lines, dem_in.temp_bins, nwalkers=20, nsteps=100)

# Get results
samples = dem_result.samples

fig, ax = plt.subplots()
# Plot the last guess from the last step of each walker used in the MCMC run
dem_result.plot_final_samples(ax)
# Plot 'true' input DEM
ax.scatter(dem_in.temp_bins.bin_centers, dem_in.values, label="Input DEM")

ax.set_yscale("log")
ax.legend()

fig, ax = plt.subplots()
ax.plot(-dem_result.sampler.get_log_prob())
ax.set_ylabel("-log(probability)")
ax.set_xlabel("Walker step")
ax.set_yscale("log")
plt.show()



from pathlib import Path
from typing import Dict

import astropy.units as u
import xarray as xr

from demcmc.emission import ContFuncDiscrete

__all__ = ["load_cont_funcs"]


def load_cont_funcs(path: Path) -> Dict[str, ContFuncDiscrete]:
    """
    Load a set of contribution functions from a netCDF file.

    The file should contain a 2D array of data, with coordinates named
    "Temperature" for the temperatures and "Line" for the line names.

    Parameters
    ----------
    path : os.PathLike
        Path of the netCDF file.

    Returns
    -------
    dict[str, ContFuncDiscrete]
        Mapping of line name to the loaded contribution function.
    """
    da = xr.open_dataarray(path)
    temps = da.coords["Temperature"].values
    lines = da.coords["Line"].values

    cont_funcs = {}
    for line in lines:
        cont_funcs[line] = ContFuncDiscrete(
            temps=temps * u.K,
            values=da.loc[line, :].values * u.cm**5 / u.K,
            name=line,
        )

    return cont_funcs



"""
Tutorial
========
This page contains a tutorial, stepping the user through estimating a DEM.
In this tutorial we use a single pixel of real data observed by Hinode/EIS.
"""

######################################################################################
# Start by importing the required modules

import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
from astropy.visualization import quantity_support

from demcmc import (
    EmissionLine,
    TempBins,
    load_cont_funcs,
    plot_emission_loci,
    predict_dem_emcee,
)
from demcmc.sample_data import fetch_sample_data

quantity_support()

######################################################################################
# Load the paths to the sample intensity and contribution function data
intensity_path, cont_func_path = fetch_sample_data()

######################################################################################
# To start with we'll load a set of line intensities. These have been taken from a
# single pixel of a fitted Hinode/EIS intensity map, and saved into a netCDF file for
# easy loading.
line_intensities = xr.open_dataarray(intensity_path)
print(line_intensities)

######################################################################################
# Now we've loaded the intensities, lets do a quick visualisation
fig, ax = plt.subplots(constrained_layout=True)
ax.barh(
    line_intensities.coords["Line"],
    line_intensities.loc[:, "Intensity"],
)
ax.set_xlim(0)
ax.set_xlabel("Observed intensity")

######################################################################################
# As well as observed intensities, the second ingredient we need for estimating a DEM
# is the theoretical contribution functions for each observed line. These have been
# pre-computed and included as sample data. The `load_cont_funcs` function
# provides functionality to load these from the saved netCDF file.

cont_funcs = load_cont_funcs(cont_func_path)
for line in cont_funcs:
    print(f"{line}: {cont_funcs[line]}")

######################################################################################
# ``cont_funcs`` is a dictionary that maps the emission line to a contribution function
# object. This object stores a pre-computed contribution function at a range of
# discrete temperature values. Lets do a visualisation of all the contribution
# functions:
fig, ax = plt.subplots()
for line in cont_funcs:
    ax.plot(cont_funcs[line].temps, cont_funcs[line].values, label=line)

ax.set_xscale("log")
ax.set_yscale("log")
ax.legend()
ax.set_title("Contribution functions")

######################################################################################
# Having loaded intensities and contribution functions for each emission line, the
# final preparation step we need to do is create a list `EmissionLine` objects for
# each line. This object stores both the observations and theoretical contribution
# function in a single place for each line.
lines = []
for line in line_intensities.coords["Line"].values:
    if "Fe" not in line:
        # Only use iron lines in the DEM inversion
        continue

    cont_func = cont_funcs[line]
    intensity = line_intensities.loc[line, :]

    line = EmissionLine(
        cont_func,
        intensity_obs=intensity.loc["Intensity"].values,
        sigma_intensity_obs=intensity.loc["Error"].values,
    )
    lines.append(line)

######################################################################################
# Now we can run the DEM inversion! We have to decide what temperature bins we want
# to estimate the DEM in. For this example we'll use 12 bins, with a width of 0.1 MK.
#
# ``demcmc`` uses the ``emcee``` package to run the MCMC sampler, and attempt to find
# the best values of the DEM that match the line intensity observations.

temp_bins = TempBins(10 ** np.arange(5.6, 6.8, 0.1) * u.K)
# Run DEM inversion
dem_result = predict_dem_emcee(lines, temp_bins, nwalkers=50, nsteps=100)
print(dem_result)

######################################################################################
# The `DEMOutput` class has several useful properties. One of those is ``.samples``,
# which returns the samples of the DEM values at the final step of each MCMC walker.
print(dem_result.samples.shape)
print(dem_result.samples)

######################################################################################
# To visualise the result, we can plot the final samples, and then plot the emission
# loci on the same axes. The emission loci give an upper bound on the DEM.

fig, ax = plt.subplots()

dem_result.plot_final_samples(ax)
plot_emission_loci(lines, ax, color="tab:blue")

ax.set_xscale("log")
ax.set_yscale("log")
ax.set_xlim(temp_bins.min * 0.9, temp_bins.max * 1.1)
ax.set_ylim(1e18 * u.cm**-5, 1e26 * u.cm**-5)
ax.set_title("Emission loci (blue) and DEM estimates (black)")

######################################################################################
# Finally, we should check if the MCMC walker converged or not.
chain = dem_result.sampler.get_chain()
nsamplers = chain.shape[1]
nparams = chain.shape[2]

fig, axs = plt.subplots(nrows=nparams, sharex=True, figsize=(6, 20))

for ax, param in zip(axs, range(chain.shape[2])):
    for sampler in range(chain.shape[1]):
        ax.plot(chain[:, sampler, param], color="tab:blue", alpha=0.1)

    # Plot average of each walker at each step
    ax.plot(np.mean(chain[:, :, param], axis=1), color="k")

    ax.set_yscale("log")
    ax.margins(x=0)
    ax.xaxis.grid()

fig.subplots_adjust(hspace=0)
axs[0].set_title("Parameter estimates as a function of MCMC step")

######################################################################################
# There are still large scale variations going on for each parameter, so clearly the
# MCMC run has not converged. When running ``demcmc`` yourself make sure to set an
# appropriate number of walkers and steps so the parameter estimates converge!

plt.show()



"""
Functions for carrying out MCMC estimation of DEMs.
"""
from typing import List, Sequence, Tuple

import emcee
import numpy as np

from demcmc.dem import DEMOutput, TempBins
from demcmc.emission import EmissionLine

__all__ = ["predict_dem_emcee"]


def _log_prob_line(
    line: EmissionLine,
    temp_bins: TempBins,
    dem_guess: np.ndarray,
) -> float:
    """
    Get log probability of intensity stored in ``line`` for the given DEM.

    Returns
    -------
    float
        Probability.
    """
    intensity_pred = line._I_pred(temp_bins, dem_guess)
    ret = -float(
        ((line.intensity_obs - intensity_pred) / line.sigma_intensity_obs) ** 2
    )
    return ret


def _log_prob_lines(
    lines: List[EmissionLine],
    temp_bins: TempBins,
    dem_guess: np.ndarray,
) -> float:
    """
    Get log probability of all line intensities stored in ``lines`` for the given DEM.

    Returns
    -------
    float
        Probability.
    """
    probbs = [_log_prob_line(line, temp_bins, dem_guess) for line in lines]
    return float(np.sum(probbs))


def _log_prob_single_variation(
    dem_val: float,
    idx_varied: int,
    dem_guess: np.ndarray,
    temp_bins: TempBins,
    lines: List[EmissionLine],
) -> float:
    """
    log probability of a given set of DEM values, varying one of them.
    The DEM values are passed as logs to enforce positivity.]

    Parameter
    ---------
    dem_val :
        DEM value being varied.
    idx_varied :
        Index of dem_guess that is being varied.
    dem_guess :
        Rest of the DEM values.

    Returns
    -------
    float
        Probability.
    """
    dem_guess[idx_varied] = dem_val
    return _log_prob(dem_guess, temp_bins, lines)


def _log_prob(
    dem_guess: np.ndarray,
    temp_bins: TempBins,
    lines: List[EmissionLine],
) -> float:
    """
    log probability of a given set of DEM values, varying one of them.
    The DEM values are passed as logs to enforce positivity.]

    Parameter
    ---------
    dem_val :
        DEM value being varied.
    dem_guess :
        Rest of the DEM values.

    Returns
    -------
    float
        Probability.
    """
    if np.any(dem_guess < 0):
        return float(-np.inf)

    p = _log_prob_lines(lines, temp_bins, dem_guess)
    return p


def predict_dem_emcee(
    lines: Sequence[EmissionLine],
    temp_bins: TempBins,
    *,
    nsteps: int,
    nwalkers: int,
    progress: bool = True,
) -> DEMOutput:
    """
    Estimate DEM from a number of emission lines.

    Parameters
    ----------
    lines : Sequence[EmissionLine]
        Emission lines.
    temp_bins : TempBins
        Temperature bins to predict DEM in.
    nsteps : int
        Total number of steps for the MCMC walkers to take.
    nwalkers : int
        Number of MCMC walkers to use.
    progress : bool
        Whether to show a progress bar for the MCMC walking.

    Returns
    -------
    DEMOutput
        Output container.
    """
    # Initial DEM value guesses
    n_dem = len(temp_bins)
    dem_guess = 1e22 * np.ones(n_dem)

    # Start by running emcee on each of the parameters individually
    #
    # This speeds up getting started because there instead of searching
    # an N-dimensional space, the search is done on N 1-dimensional
    # spaces.
    dem_guess, _ = _vary_values_independently(lines, temp_bins, dem_guess, nsteps=100)

    # Now run MCMC across the ful N-dimensional space to get the final guess
    dem_guess = np.repeat(np.atleast_2d(dem_guess), nwalkers, axis=0)
    dem_guess += np.random.rand(*dem_guess.shape) * 0.01 * dem_guess

    sampler = emcee.EnsembleSampler(nwalkers, n_dem, _log_prob, args=[temp_bins, lines])
    sampler.run_mcmc(dem_guess, nsteps, progress=progress)

    return DEMOutput._from_sampler(sampler, temp_bins)


def _vary_values_independently(
    lines: Sequence[EmissionLine],
    temp_bins: TempBins,
    dem_guess: np.ndarray,
    *,
    nsteps: int,
) -> Tuple[np.ndarray, List[emcee.EnsembleSampler]]:
    ndim = 1
    nwalkers = 3
    n_dem = len(temp_bins)

    parameter_guess = np.repeat(np.atleast_2d(dem_guess), nwalkers, axis=0)
    # Add randomness to initial guesses
    parameter_guess += np.random.rand(*parameter_guess.shape) * 0.01 * parameter_guess

    samplers = []
    for i in range(n_dem):
        sampler = emcee.EnsembleSampler(
            nwalkers,
            ndim,
            _log_prob_single_variation,
            args=[i, dem_guess, temp_bins, lines],
        )
        # Run sampler
        sampler.run_mcmc(parameter_guess[:, i : i + 1], nsteps=nsteps, progress=False)
        samplers.append(sampler)

        samples = sampler.get_chain()
        # Take mean of the last 10 steps
        dem_guess[i] = np.mean(samples[-10:, :, :])
        parameter_guess[:, i] = samples[-1, :, 0]

    return dem_guess, samplers



"""
Load sample data.
"""
from pathlib import Path
from typing import Tuple

import pooch

REGISTRY = pooch.create(
    path=pooch.os_cache("demcmc"),
    base_url="doi:10.5281/zenodo.7534284/",
    registry={
        "sample_cont_func.nc": "md5:6bf547459d83d24b0da2736f36380afb",
        "sample_intensity_values.nc": "md5:f8c9481705cd0295cf1ad0f4fa41caf5",
    },
)


def fetch_sample_data() -> Tuple[Path, Path]:
    return REGISTRY.fetch("sample_intensity_values.nc"), REGISTRY.fetch(
        "sample_cont_func.nc"
    )



"""
Classes for working with DEM data.
"""
from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Any, Iterator, Optional, Tuple

import astropy.units as u
import emcee
import numpy as np
import xarray as xr
from matplotlib.axes import Axes

from demcmc.units import u_dem, u_temp

__all__ = ["BinnedDEM", "TempBins", "DEMOutput"]


class TempBins:
    """
    A set of temperature bins.

    The bins are defined through the bin edges.

    Parameters
    ----------
    edges : astropy.units.Quantity
        Bin edges.
    """

    @u.quantity_input(edges=u_temp)
    def __init__(self, edges: u.Quantity):
        self._edges = edges
        self._edges_arr = edges.to_value(u_temp)
        self._hash = id(self)

    @property
    def edges(self) -> u.Quantity[u_temp]:
        """
        Edges of the temperature bins.
        """
        return self._edges

    @edges.setter
    def edges(self, val: Any) -> None:
        """
        Raises an error.
        """
        raise RuntimeError("ContFuncDiscrete instances are immutable")

    @property
    def min(self) -> u.Quantity[u.K]:
        """
        Lower bound of the temperature bins.
        """
        return self.edges.min()

    @property
    def max(self) -> u.Quantity[u.K]:
        """
        Upper bound of the temperature bins.
        """
        return self.edges.max()

    def __hash__(self) -> int:
        return self._hash

    @cached_property
    def bin_widths(self) -> u.Quantity:
        """
        Widths of the bins.
        """
        return np.diff(self.edges)

    @cached_property
    def _bin_widths_arr(self) -> np.ndarray:
        """
        Widths of the bins as a bare array.
        """
        return np.diff(self._edges_arr)

    @cached_property
    def bin_centers(self) -> u.Quantity:
        """
        Centers of the bins.
        """
        return (self.edges[:-1] + self.edges[1:]) / 2

    @cached_property
    def _bin_centers_MK(self) -> np.ndarray:
        """
        Centers of the bins as bare numbers in units of MK.
        """
        return self.bin_centers.to_value(u.MK)

    def __len__(self) -> int:
        """
        Number of bins.
        """
        # int is just to make mypy happy
        return int(self.edges.size - 1)

    def iter_bins(self) -> Iterator[Tuple[u.Quantity, u.Quantity]]:
        """
        Iterate through lower/upper bounds of temperature bins.

        Yields
        ------
        lower_edge : astropy.units.Quantity
            Lower edge of bin.
        upper_edge : astropy.units.Quantity
            Upper edge of bin.
        """
        for i in range(len(self)):
            yield self.edges[i], self.edges[i + 1]


@dataclass
class BinnedDEM:
    """
    A DEM binned over a range of temperature values.

    Parameters
    ----------
    temp_bins : TempBins
        Temperature bins.
    values : astropy.units.Quantity
        DEM values.
    """

    temp_bins: TempBins
    values: u.Quantity

    def __init__(self, temp_bins: TempBins, values: u.Quantity[u.cm**-5]):
        self.temp_bins = temp_bins
        self.values = values

        self._values_arr = self.values.to_value(u_dem)


class DEMOutput:
    """
    Output from running DEM calculation.

    This is not intended to be created by users.
    """

    _sampler: Optional[emcee.EnsembleSampler]
    _samples: u.Quantity
    _temp_bins: TempBins

    @property
    def sampler(self) -> emcee.EnsembleSampler:
        """
        Sampler used by emcee to calculated the DEM.
        """
        return self._sampler

    @property
    def temp_bins(self) -> TempBins:
        """
        Temperature bins.
        """
        return self._temp_bins

    @property
    def samples(self) -> u.Quantity:
        """
        Return the last set of samples from the walker.
        """
        return self._samples

    @property
    def n_samples(self) -> int:
        """
        Number of walkers used (same as number of final samples).
        """
        return int(self.samples.shape[0])

    def iter_binned_dems(self) -> Iterator[BinnedDEM]:
        """
        Iterate across each final sample, returning a `BinnedDEM`.

        Yields
        ------
        BinnedDEM
        """
        for i in range(self.n_samples):
            yield BinnedDEM(self.temp_bins, self.samples[i, :])

    def plot_final_samples(self, ax: Axes, **kwargs: Any) -> None:
        """
        Plot the final samples of the MCMC walker.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes to plot the samples on.
        **kwargs : dict
            Keyword arguments are passed to :meth:`~matplotlib.axes.Axes.stairs`.
        """
        kwargs.setdefault("color", "k")
        kwargs.setdefault("alpha", 0.1)
        kwargs.setdefault("linewidth", 1)
        for i in range(self.samples.shape[0]):
            ax.stairs(self.samples[i, :], self.temp_bins.edges, **kwargs)

    def save(self, path: Path) -> None:
        """
        Save a computed DEM to a netCDF file.

        Parameters
        ----------
        path : pathlib.Path
            Path to save to. Should end in ``".nc"``.
        """
        temp_centers = self.temp_bins.bin_centers
        temp_edges = self.temp_bins.edges

        samplers = np.arange(self.samples.shape[0])
        coords = {"Sampler": samplers, "Temp bin center": temp_centers.to_value(u_temp)}

        da = xr.DataArray(
            data=self.samples,
            coords=coords,
            attrs={"Temp bin edges": temp_edges.to_value(u_temp)},
        )
        da.to_netcdf(path)

    @classmethod
    def load(cls, path: Path) -> DEMOutput:
        """
        Load a computed DEM from a netCDF file.

        This will load files saved by the ``.save()`` method.

        Parameters
        ----------
        path : pathlib.Path
            Path to netCDF file.

        Returns
        -------
        DEMOutput
        """
        da = xr.load_dataarray(path)
        self = cls()
        self._temp_bins = TempBins(da.attrs["Temp bin edges"] * u_temp)
        self._samples = da.data * u_dem
        return self

    @classmethod
    def _from_sampler(
        cls, sampler: emcee.EnsembleSampler, temp_bins: TempBins
    ) -> DEMOutput:
        self = cls()
        self._sampler = sampler
        self._temp_bins = temp_bins
        self._samples = self.sampler.get_chain()[-1, :, :] * u_dem
        return self



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "demcmc"
copyright = "2022, David Stansby"
author = "David Stansby"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

default_role = "any"

extensions = [
    "matplotlib.sphinxext.plot_directive",
    "myst_parser",
    "numpydoc",
    "sphinx.ext.graphviz",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx_gallery.gen_gallery",
]

exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "examples/*"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"

# -- Extension configuration
sphinx_gallery_conf = {
    "examples_dirs": "./tutorials",  # path to your example scripts
    "gallery_dirs": "./_auto_examples",  # path to where to save gallery generated output
    "download_all_examples": False,
    "capture_repr": (),
}

numpydoc_show_class_members = False
numpydoc_xref_param_type = True

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "astropy": ("https://docs.astropy.org/en/stable/", None),
    "matplotlib": ("https://matplotlib.org/stable", None),
    "emcee": ("https://emcee.readthedocs.io/en/stable/", None),
}

numpydoc_validation_checks = {
    "all",
    "SA01",  # Allow omitting See Also section
    "EX01",  # Allow omitting Examples section
    "ES01",  # Allow omitting extended summary section
}

autodoc_default_options = {
    "members": True,
}



from demcmc.dem import DEMOutput, TempBins
from demcmc.emission import ContFuncDiscrete, EmissionLine, plot_emission_loci
from demcmc.io import load_cont_funcs
from demcmc.mcmc import predict_dem_emcee
